////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_CONDITIONAL_END_H_INC_
#define REFLIB_CONDITIONAL_END_H_INC_

#include <boost/mpl/aux_/has_size.hpp>

// VC6 has significant problems with member pointers as template arguments - 
// hence detection requires here not only a done member function but also
// a DoneTag typedef - sorry for any inconvenience :-)
namespace refl
{
    // attribute traversions might want to stop before the last
    // attribute is handled (for instance the attributewise comparison
    // of reflected classes may know the result after comparing the 
    // first couple of attributes).
    // To use this your Attribute Function object most have a  
    // const or non-const member function bool done(). 
    struct CheckEndCondition{};
    struct DontCheckEndCondition{};
    template<int> struct EnumToEndConditionAnswer
    {
        typedef CheckEndCondition type;
    };
    template<> struct EnumToEndConditionAnswer<0>
    {
        typedef DontCheckEndCondition type;
    };
    
    BOOST_MPL_HAS_XXX_TRAIT_DEF(DoneTag)

    template<class Type> struct HasConditionalEnd
    {
        enum {value = has_DoneTag<Type>::value}; 
        typedef typename EnumToEndConditionAnswer<value>::type type;
    };
}
#endif
