////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_ISRUNTIMEITERABLE_H_INC_
#define REFLIB_ISRUNTIMEITERABLE_H_INC_

#include <boost/type_traits/remove_cv.hpp>

/*  too expensive - clients that need to know that an instance of an 
    std::container is runtime iterable should have #included the corresponding 
    header.
#include <list>
#include <vector>
#include <set>
#include <map>
#include <deque>
*/

namespace std
{
    template <class T, class Allocator> class vector;
    
    template <class T, class Allocator> class list;
    
    template <class T, class Allocator> class deque;

    template <class T, class Compare, class Allocator> class set;
    
    template <class T, class Compare, class Allocator> class multiset;

    template <class Key, class T, class Compare, class Allocator> class map;

    template <class Key, class T, class Compare,class Allocator> class multimap;
}


struct RunTIterableObjectTag{};

struct NotRunTIterableObjectTag{};

namespace refl
{
    namespace detail
    {
        template <class T> struct IsRuntimeIterable
        {
            typedef NotRunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = false);
        };

        // For any set of template arguments - an std::vector is runtime-iterable
        template <class T, class Allocator> struct IsRuntimeIterable<std::vector<T,
                                                                    Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::deque is runtime-iterable
        template <class T, class Allocator>
        struct IsRuntimeIterable<std::deque<T, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::list is runtime-iterable
        template <class T, class Allocator>
        struct IsRuntimeIterable<std::list<T, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::set is runtime-iterable
        template <class T, class Compare, class Allocator>
        struct IsRuntimeIterable<std::set<T, Compare, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::multiset is runtime-iterable
        template <class T, class Compare, class Allocator>
        struct IsRuntimeIterable<std::multiset<T, Compare, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::map is runtime-iterable
        template <class Key, class T, class Compare, class Allocator>
        struct IsRuntimeIterable<std::map<Key, T, Compare, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };

        // For any set of template arguments - an std::multimap is runtime-iterable
        template <class Key, class T, class Compare, class Allocator>
        struct IsRuntimeIterable<std::multimap<Key, T, Compare, Allocator> >
        {
            typedef RunTIterableObjectTag Detect;
            BOOST_STATIC_CONSTANT(int, value = true);
        };
    }
    template <class T> struct IsRuntimeIterable:
    detail::IsRuntimeIterable<typename boost::remove_cv<T>::type>
    {};
}
#endif //
