////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_ISRUNTIMEITERABLE_H_INC_
#define REFLIB_ISRUNTIMEITERABLE_H_INC_

/*  too expensive - clients that need to know that an instance of an 
    std::container is runtime iterable should have #included the corresponding 
    header.
#include <list>
#include <vector>
#include <set>
#include <map>
#include <deque>
*/

namespace std
{
    template <class T, class Allocator> class vector;
    
    template <class T, class Allocator> class list;
    
    template <class T, class Allocator> class deque;

    template <class T, class Compare, class Allocator> class set;
    
    template <class T, class Compare, class Allocator> class multiset;

    template <class Key, class T, class Compare, class Allocator> class map;

    template <class Key, class T, class Compare,class Allocator> class multimap;
}

struct RunTIterableObjectTag{};

struct NotRunTIterableObjectTag{};

namespace refl
{
    /*
        See the machinery in Andrei Alexandrescu "Modern C++ Design" 2.7 Detecting 
        Convertability" who the hack knows why gcc 3.2 nags about "use of undefined 
        type IsRuntimeIterable" with this implementation
    */
    namespace detail
    {
        template<int i> struct Int2ContainerAnswer
        {
            typedef RunTIterableObjectTag result;
        };

        template<> struct Int2ContainerAnswer<0>
        {
            typedef NotRunTIterableObjectTag result;
        };
    }
    template <class ClassToCheck> class IsRuntimeIterable
    {
        typedef char YesType;
        struct NoType{ char makeThisBiggerAsYesType[256];};
        template <class T, class Allocator>
            static YesType check(std::vector<T,Allocator>);
        template <class T, class Allocator>
            static YesType check(std::list<T,Allocator>);
        template <class T, class Allocator>
            static YesType check(std::deque<T,Allocator>);
        template <class T, class Compare, class Allocator>
            static YesType check(std::set<T, Compare, Allocator>);
        template <class T, class Compare, class Allocator>
            static YesType check(std::multiset<T, Compare, Allocator>);
        template <class Key, class T, class Compare, class Allocator>
            static YesType check(std::map<Key, T, Compare, Allocator>);
        template <class Key, class T, class Compare, class Allocator>
            static YesType check(std::multimap<Key, T, Compare, Allocator>);

        static NoType check(...);

        // in case that the default ctor of ClassToCheck is not accessible.
        static ClassToCheck makeClassToCheck();
    public: 
        enum { value = sizeof(check(makeClassToCheck())) == sizeof(YesType)};
        typedef typename detail::Int2ContainerAnswer<value>::result Detect;
    };
}
#endif //
