////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_LEXICOGRAPHICALSORT_H_INC_
#define REFLIB_LEXICOGRAPHICALSORT_H_INC_

template<class FirstSortFunction, class SecondSortFunction> 
class LexicographicalSort
{
public:
    LexicographicalSort(FirstSortFunction, SecondSortFunction);
    template<class ItemType> bool operator()(const ItemType& lhs, 
                                            const ItemType& rhs)const
    {
        return firstSortFunction_(lhs,rhs) ||
                (!firstSortFunction_(rhs,lhs) && secondSortFunction_(lhs,rhs));
        /* slightly slower but more readable:
        if(firstSortFunction_(lhs,rhs))
            return true;
        else
        {
            if(firstSortFunction_(rhs,lhs))
                return false;
            else    // neither lhs precedes rhs nor rhs precedes lhs with 
                    // respect to the firstSortFunction
                    // that is lhs and rhs are equivalent with respect to the 
                    // firstSortFunction
                return secondSortFunction_(lhs,rhs);
        }*/
    }
private:
    FirstSortFunction firstSortFunction_;
    SecondSortFunction secondSortFunction_;
};

template<class FirstSortFunction, class SecondSortFunction>
LexicographicalSort<FirstSortFunction,
SecondSortFunction>::LexicographicalSort(FirstSortFunction firstSortFunction, 
                                         SecondSortFunction secondSortFunction):
firstSortFunction_(firstSortFunction),secondSortFunction_(secondSortFunction)
{}


template<class FirstSortFunction, class SecondSortFunction> 
LexicographicalSort<FirstSortFunction, SecondSortFunction>
makeLexicographicalSort(FirstSortFunction fun1, SecondSortFunction thisIsfunToo)
{
    return LexicographicalSort<FirstSortFunction,SecondSortFunction>(fun1,
                                                                thisIsfunToo);
}
#endif      

