////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_BASECLASSITERRT_FILTER_H_INC_
#define REFLIB_BASECLASSITERRT_FILTER_H_INC_

#include <typeinfo>
#include <set>
#include <Loki/Typelist.h>
#include <Loki/LokiTypeInfo.h>

#include "../qualifiedtype.h"

#include <boost/pending/ct_if.hpp>
#include "boost/type_traits/is_same.hpp"

#include "passtype.h"

namespace refl
{
    typedef std::set<Loki::TypeInfo> BaseClassLog;

    struct VirtualTag{};
    
    struct NotVirtualTag{};

    // for iterations that evaluate class information like:
    // which members does a class have?
    struct EvaluateTagsOnly{};
    
    struct EvaluateMembers{};

    template<int> struct IsVirtual
    {
        typedef VirtualTag type;
    };
    
    template<> struct IsVirtual<0>
    {
        typedef NotVirtualTag type;
    };

    inline bool vbaseClassNotHandled(const BaseClassLog& baseClasses,
                                        Loki::TypeInfo baseClassID,VirtualTag)
    {
        return baseClasses.find(baseClassID) == baseClasses.end();
    }

    inline bool vbaseClassNotHandled(const BaseClassLog&,Loki::TypeInfo ,
                                    NotVirtualTag)
    {
        return true;
    }

#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
    inline int insertHandledVBaseClass(BaseClassLog& baseClasses,
                                      Loki::TypeInfo baseClassID,VirtualTag)
    {
        baseClasses.insert(baseClassID);
        return 1;
    }
    inline int insertHandledVBaseClass(const BaseClassLog& baseClasses,
                                                Loki::TypeInfo ,NotVirtualTag)
    {
        return 1;
    }
#else
    inline void insertHandledVBaseClass(BaseClassLog& baseClasses,
                                      Loki::TypeInfo baseClassID,VirtualTag)
    {
        baseClasses.insert(baseClassID);
    }
    inline void insertHandledVBaseClass(const BaseClassLog& ,
                                        Loki::TypeInfo ,NotVirtualTag)
    {}
#endif
                        

    template<class ReflectedClass> struct AllBaseClasses
    {
        enum {TotalNumBaseClasses = ReflectedClass::NumBaseClasses};
        typedef typename ReflectedClass::BaseClasses BaseClassList;
        template<int Num> struct BaseClassVisitor
        {
            template<class CTRecursion> static void processSingleBaseClass(
                                            ReflectedClass& recordData,
                                            CTRecursion& ctRecursion, 
                                            BaseClassLog& handledVBases)
            {
                typedef typename ReflectedClass::template GetBaseClass<Num,
                    ReflectedClass::ThisClass> 
                    CurrentBaseClassTraits;
                typedef typename CurrentBaseClassTraits::type 
                    CurrentBaseClassType;
                // apply const if the context is const
                typedef typename GetQualifiedMemberType<ReflectedClass,
                                            CurrentBaseClassType>::type 
                                            BaseClass;
                typedef typename 
                    IsVirtual<CurrentBaseClassTraits::isVirtual>::type VTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                // this cast from derived to base is necessary because this way 
                // I can pass the type of the baseclass to processAsBaseClass 
                static const bool notHandledYet = vbaseClassNotHandled(
                        handledVBases,Loki::TypeInfo(typeid(BaseClass)),VTag());
                if(notHandledYet)
                    processAsBaseClass(static_cast<BaseClass&>(recordData),
                                ctRecursion, handledVBases, VTag());
                // more than once is more then one time too often
            } 
            // binary overload
            template<class CTRecursion> static void 
                processSingleBaseClass(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                       CTRecursion& ctRecursion, BaseClassLog& handledVBases)
            {
                typedef typename ReflectedClass::template GetBaseClass<Num,
                    ReflectedClass::ThisClass> 
                    CurrentBaseClassTraits;
                typedef typename CurrentBaseClassTraits::type 
                    CurrentBaseClassType;
                // apply const if the context is const
                typedef typename GetQualifiedMemberType<ReflectedClass,
                                            CurrentBaseClassType>::type 
                                            BaseClass;
                typedef typename 
                    IsVirtual<CurrentBaseClassTraits::isVirtual>::type VTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                // this cast from derived to base is necessary because this way 
                // I can pass the type of the baseclass to processAsBaseClass 
                static const bool notHandledYet = vbaseClassNotHandled(
                        handledVBases,Loki::TypeInfo(typeid(BaseClass)),VTag());
                if(notHandledYet)
                    processAsBaseClass(static_cast<BaseClass&>(lhsRecord),
                                static_cast<BaseClass&>(rhsRecord),
                                ctRecursion, handledVBases, VTag());
                // more than once is more then one time too often
            } 
            // if CTRecursion is inclined to iterate over static members or tags only
            template<class CTRecursion, class TagsOrMembers> static void processSingleBaseClass(
                CTRecursion& ctRecursion, BaseClassLog& handledVBases,TagsOrMembers tagsOrMembers)
            {
                // typedefs at class scope that refer a dependent member 
                // template cause compiler errors with VC++ 6.0
                typedef typename ReflectedClass::template GetBaseClass<Num,
                    ReflectedClass::ThisClass> 
                    CurrentBaseClassTraits;
                typedef typename CurrentBaseClassTraits::type 
                    CurrentBaseClassType;
                // apply const if the context is const
                typedef typename GetQualifiedMemberType<ReflectedClass,
                                            CurrentBaseClassType>::type 
                                            BaseClass;
                typedef typename 
                    IsVirtual<CurrentBaseClassTraits::isVirtual>::type VTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                static const bool notHandledYet = vbaseClassNotHandled(
                    handledVBases,Loki::TypeInfo(typeid(BaseClass)),VTag());
                if(notHandledYet)
                    processAsBaseClass(ctRecursion,handledVBases,
                                    PassType<BaseClass>(), VTag(), tagsOrMembers);
                // more than once is more then one time too often
            } 
            // there can only be one type of baseclass in this instantiation of 
            // BaseClassVisitor - unfortunately that type is not available at 
            // class scope because MSVC++ 6.0 tries to instantiate the needed 
            // typedefs without having seen the template arguments and then 
            // chokes.
            // Similarly I do pass the virtualness of that base class as Tag 
            // because this avoids repeating the typedefs that "calculate" that 
            // tag.
            // Note that this does not generate any additional code-bloat 
            // because that Tag depends uniquely on the position in the 
            // baseclass graph which maps uniquely to *this instantiation of 
            // BaseClassVisitor.
            template<class BaseClass, class CTRecursion, class VTag> static void
                processAsBaseClass(BaseClass& recordData, 
                              CTRecursion& ctRecursion, 
                              BaseClassLog& handledVBases, VTag vTag)    
            {  
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
                // The traversion graph depends on static structure only.
                // Hence it suffices to record that information once.
                // Having done that, the call to insert will never be exectuted 
                // any more becuase it is part of the initializer statement
                // for the static storage of "withMyContribution".
                // The compiler generates code so that this storage will be 
                // initialized only once.
                // Now comes the hack - we store the return of insertHandledVBaseClass
                // because we are only interested in the side-effect that this
                // function will be only executed once.
                static const int executeOnce = 
                    insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#else
                insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#endif
                typedef AllBaseClasses<BaseClass> BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(recordData,ctRecursion,
                                                handledVBases);
                ctRecursion(recordData);
            }
            // binary overload
            template<class BaseClass, class CTRecursion, class VTag> static void
                processAsBaseClass(BaseClass& lhsRecord, 
                              BaseClass& rhsRecord,CTRecursion& ctRecursion, 
                              BaseClassLog& handledVBases, VTag vTag)    
            {  
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
                // The traversion graph depends on static structure only.
                // Hence it suffices to record that information once.
                // Having done that, the call to insert will never be exectuted 
                // any more becuase it is part of the initializer statement
                // for the static storage of "withMyContribution".
                // The compiler generates code so that this storage will be 
                // initialized only once.
                // Now comes the hack - we store the return of insertHandledVBaseClass
                // because we are only interested in the side-effect that this
                // function will be only executed once.
                static const int executeOnce = 
                    insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#else
                insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#endif
                typedef AllBaseClasses<BaseClass> BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(lhsRecord, rhsRecord,ctRecursion,
                                                handledVBases);
                ctRecursion(lhsRecord, rhsRecord);
            }
            template<class BaseClass,class CTRecursion, class VTag, class TagsOrMembers> 
            static void processAsBaseClass(CTRecursion& ctRecursion, 
                            BaseClassLog& handledVBases, 
                            PassType<BaseClass> baseClassType, VTag vTag,TagsOrMembers tagsOrMembers)    
            {  
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
                // The traversion graph depends on static structure only.
                // Hence it suffices to record that information once.
                // Having done that, the call to insert will never be exectuted 
                // any more becuase it is part of the initializer statement
                // for the static storage of "withMyContribution".
                // The compiler generates code so that this storage will be 
                // initialized only once.
                // Now comes the hack - we store the return of insertHandledVBaseClass
                // because we are only interested in the side-effect that this
                // function will be only executed once.
                static const int executeOnce = 
                    insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#else
                insertHandledVBaseClass(handledVBases,
                                        Loki::TypeInfo(typeid(BaseClass)),vTag);
#endif
                processAsBaseClass(ctRecursion,handledVBases,baseClassType,tagsOrMembers);
            }
            template<class BaseClass,class CTRecursion> 
            static void processAsBaseClass(CTRecursion& ctRecursion, 
                            BaseClassLog& handledVBases, 
                            PassType<BaseClass> baseClassType,EvaluateMembers)    

            {
                typedef AllBaseClasses<BaseClass> BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(ctRecursion,
                                                handledVBases);
                ctRecursion(baseClassType,ClassVariableTag());
            }
            template<class BaseClass,class CTRecursion> 
            static void processAsBaseClass(CTRecursion& ctRecursion, 
                            BaseClassLog& handledVBases, 
                            PassType<BaseClass> baseClassType,EvaluateTagsOnly tagsOnly)    

            {
                typedef AllBaseClasses<BaseClass> BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(ctRecursion,
                                                handledVBases,
                                                tagsOnly);
                ctRecursion(baseClassType);
            }
        };
        template<int RestToDo> struct HorizontalIter:
                    BaseClassVisitor<TotalNumBaseClasses-RestToDo>
        {
            template<class CTRecursion> static void process(
                                ReflectedClass& recordData,
                                CTRecursion& ctRecursion, 
                                BaseClassLog& handledVBases)
            {
                processSingleBaseClass(recordData, ctRecursion,handledVBases);
                HorizontalIter<RestToDo-1>::process(recordData, 
                                                     ctRecursion,handledVBases);
            }
            template<class CTRecursion> static void 
                process(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                      CTRecursion& ctRecursion, BaseClassLog& handledVBases)
            {
                processSingleBaseClass(lhsRecord, rhsRecord, ctRecursion,handledVBases);
                HorizontalIter<RestToDo-1>::process(lhsRecord, rhsRecord, 
                                                     ctRecursion,handledVBases);
            }
            template<class CTRecursion, class TagsOrMembers> 
            static void process(CTRecursion& ctRecursion, 
                                BaseClassLog& handledVBases,
                                TagsOrMembers tagsOrMembers)
            {
                processSingleBaseClass(ctRecursion,handledVBases, tagsOrMembers);
                HorizontalIter<RestToDo-1>::process(ctRecursion,handledVBases, tagsOrMembers);
            }
        };

        template<> struct HorizontalIter<1>:
                            BaseClassVisitor<TotalNumBaseClasses-1>
        {
            template<class CTRecursion> static void process(
                                ReflectedClass& recordData,
                                CTRecursion& ctRecursion, 
                                BaseClassLog& handledVBases)
            {
                processSingleBaseClass(recordData, ctRecursion,handledVBases);
            }
            template<class CTRecursion> static void process(
                                ReflectedClass& lhsRecord,
                                ReflectedClass& rhsRecord,
                                CTRecursion& ctRecursion, 
                                BaseClassLog& handledVBases)
            {
                processSingleBaseClass(lhsRecord, rhsRecord, ctRecursion,handledVBases);
            }
            template<class CTRecursion, class TagsOrMembers> static void 
            process(CTRecursion& ctRecursion, BaseClassLog& handledVBases, TagsOrMembers tagsOrMembers)
            {
                processSingleBaseClass(ctRecursion,handledVBases,tagsOrMembers);
            }
        };
        template<> struct HorizontalIter<0>
        {
            // RestToDo == 0 process will be called when 
            // TotalNumBaseClasses == 0 as well.
            template<class CTRecursion> static void 
            process(ReflectedClass& ,CTRecursion& , BaseClassLog&)
            {}
            template<class CTRecursion> static void 
            process(ReflectedClass&,ReflectedClass&,CTRecursion&, BaseClassLog&)
            {}
            template<class CTRecursion, class TagsOrMembers> static void 
            process(CTRecursion& , BaseClassLog&,TagsOrMembers)
            {}
        };
        template<class CTRecursion> static void
        doForeach(ReflectedClass& dataTuple, CTRecursion& ctRecursion, 
                BaseClassLog& handledVBases)
        {
            HorizontalIter<TotalNumBaseClasses>::process(dataTuple,
                                                        ctRecursion,
                                                        handledVBases);
        }
        template<class CTRecursion> static void
        doForeach(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, CTRecursion& ctRecursion, 
                BaseClassLog& handledVBases)
        {
            HorizontalIter<TotalNumBaseClasses>::process(lhsRecord,rhsRecord,
                                                        ctRecursion,
                                                        handledVBases);
        }
        // This form can only access static members of base classes
        // but on the upside is that you dont need an instance of ReflectedClass
        template<class CTRecursion> static void 
        doForeach(CTRecursion& ctRecursion, BaseClassLog& handledVBases)
        {
            HorizontalIter<TotalNumBaseClasses>::process(ctRecursion,
                                                        handledVBases,
                                                        EvaluateMembers());
        }

        template<class CTRecursion> static void 
        doForeach(CTRecursion& ctRecursion, BaseClassLog& handledVBases,EvaluateTagsOnly tagsOnly)
        {
            HorizontalIter<TotalNumBaseClasses>::process(ctRecursion,
                                                        handledVBases,
                                                        tagsOnly);
        }
    };
    // might as well be useful to have a foreachBaseClassAndThisClass algorithm
    /*template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClassAndThisClass(ReflectedClass& dataTuple,
                                    CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(dataTuple,
                                                            ctRecursion);
        ctRecursion(dataTuple);
    }*/
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& dataTuple, 
                        CompileTimeRecursion& ctRecursion)   
    {
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
        // after the first call to this function handledVBases
        // remains const. I did not test whether passing this 
        // reference around is in fact significantly cheaper
        // then constructing an empty set in each call of this
        // function.
        static BaseClassLog handledVBases;
#else
        BaseClassLog handledVBases;
#endif
        AllBaseClasses<ReflectedClass>::doForeach(dataTuple,
                                                ctRecursion,
                                                handledVBases);
    }
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, 
                        CompileTimeRecursion& ctRecursion)   
    {
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
        // after the first call to this function handledVBases
        // remains const. I did not test whether passing this 
        // reference around is in fact significantly cheaper
        // then constructing an empty set in each call of this
        // function.
        static BaseClassLog handledVBases;
#else
        BaseClassLog handledVBases;
#endif
        AllBaseClasses<ReflectedClass>::doForeach(lhsRecord,rhsRecord,
                                                ctRecursion,
                                                handledVBases);
    }
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(CompileTimeRecursion& ctRecursion)   
    {
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
        // after the first call to this function handledVBases
        // remains const. I did not test whether passing this 
        // reference around is in fact significantly cheaper
        // then constructing an empty set in each call of this
        // function.
        static BaseClassLog handledVBases;
#else
        BaseClassLog handledVBases;
#endif
        AllBaseClasses<ReflectedClass>::doForeach(ctRecursion,handledVBases);
    }
    // might need to able to distinguish between tags of static members
    // as well.
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(CompileTimeRecursion& ctRecursion, EvaluateTagsOnly tagsOnly)   
    {
#ifdef REFLIB_STORE_TRAVERSED_VBASES_STATIC
        // after the first call to this function handledVBases
        // remains const. I did not test whether passing this 
        // reference around is in fact significantly cheaper
        // then constructing an empty set in each call of this
        // function.
        static BaseClassLog handledVBases;
#else
        BaseClassLog handledVBases;
#endif
        AllBaseClasses<ReflectedClass>::doForeach(ctRecursion,
                                                handledVBases,tagsOnly);
    }
}

#endif
