////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_FOREACHTAG_H_INC_
#define REFLIB_FOREACHTAG_H_INC_

namespace refl
{
    struct IncludeBaseClasses{};

    struct DontIncludeBaseClasses{};
    
    namespace detail
    {
        template<class IgnoreOrIncludeBaseClasses> struct ForeachTagCall;

        template<> struct ForeachTagCall<DontIncludeBaseClasses>
        {
            template<class Function, class AttributeTag, class ContextType> static void call(Function& function, AttributeTag attributeTag, ContextType)
            {
                function(attributeTag);
            }
        };

        template<> struct ForeachTagCall<IncludeBaseClasses>
        {
            template<class Function, class AttributeTag, class ContextType> 
            static void call(Function& function, AttributeTag attributeTag, ContextType passedContext)
            {
                function(attributeTag, passedContext);
            }
        };
    }
    template<class ReflectedClass, class IgnoreOrIncludeBaseClasses = IncludeBaseClasses> struct ForeachTag
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};
        template<int RestToDo> struct Iter: detail::ForeachTagCall<IgnoreOrIncludeBaseClasses>
        {
            typedef detail::ForeachTagCall<IgnoreOrIncludeBaseClasses> Inherited;
            template<class Function> static void execute(Function& function)
            {
                // MSVC++ 6.0 does not allow these typedefs at class scope.
                // Hence accessing a these types is implemented per copy and 
                // paste of these lines
                typedef typename ReflectedClass::template 
                    GetNameTag<TotalNumAttributes-RestToDo,ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
                
                Inherited::call(function,CurrentAttributeTag(),PassType<ReflectedClass::ThisClass>());
                
                Iter<RestToDo-1>::execute(function);
            }
        };

        template<> struct Iter<1>: detail::ForeachTagCall<IgnoreOrIncludeBaseClasses>
        {
            typedef detail::ForeachTagCall<IgnoreOrIncludeBaseClasses> Inherited;
            template<class Function> static void execute(Function& function)
            {
                typedef typename ReflectedClass::template 
                    GetNameTag<TotalNumAttributes - 1,ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;

                Inherited::call(function,CurrentAttributeTag(),PassType<ReflectedClass::ThisClass>());
            }
        };
        template<class Function> static void execute(Function& function)
        {
            Iter<TotalNumAttributes>::execute(function);
        }
    };
    // to accumulate information about a record - was used to define the elementhandlers
    // for parsing generated xml-files with expat
    template<class ReflectedClass, class Function> void foreachTag(Function& function, ReflectedClass* = NULL)
    {
        ForeachTag<ReflectedClass>::execute(function);
    };

    // having this as a function object facilitates tagwise
    // iteration including baseclasses.
    template<class Function> struct ForeachTagFunction
    {
        ForeachTagFunction(Function& function):function_(function)
        {}
        template<class ReflectedClass> void 
        operator()(PassType<ReflectedClass>)const
        {
            foreachTag<ReflectedClass>(function_);
        }
        Function& function_;
    };

    // ReflectedClass is needed as part of the function type because otherwise
    // MSVC++ 6.0 is not able to link instantiations that differ in 
    // ReflectedClass correctly
    template<class ReflectedClass, class Function> 
    void foreachTagInclBaseClasses(Function& function, ReflectedClass* = NULL)
    {
        ForeachTagFunction<Function> tagFunction(function);
        foreachBaseClass<ReflectedClass>(tagFunction, EvaluateTagsOnly());
        tagFunction(PassType<ReflectedClass>());
    }
}
#endif // 

